/*
 * Collin Mulliner <collin[at]mulliner.org>
 * http://www.mulliner.org/android/
 * (C) 2012
 *
 *  License: LGPL
 */

#define _GNU_SORUCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <dlfcn.h>
#include <pthread.h>

/*
 *  files are from the android linker source
 */
#include "linker.h"
#include "linker_format.h"

/*
 * for test on x86 build using -m32 to make sure this is 32bit code
 * 
 * gcc finddlopen.c -g -o finddlo -DDEBUG -m32
 *
 */

unsigned int find_dlopen(char *filename)
{
	int fp;

	if ((fp = open(filename, O_RDONLY)) < 0) {
		perror("open");
		return 0;
	}

	unsigned int addr = 0;
	unsigned char *lnk;
	struct stat buf = {0};
	fstat(fp, &buf);
	lnk = (unsigned char*) mmap(NULL, buf.st_size, PROT_READ, MAP_PRIVATE, fp, 0);

	unsigned char *p = 0;
	int i;
	for (i = 0; i < buf.st_size; i++) {
		if ((p = strstr(lnk + i, "libdl.so")) != 0) {
			//printf("s = %x %d p %x pn %x\n", buf.st_size, buf.st_size, p, p - lnk);
			break;
		}
	}
		
	if (p != 0) {
		struct soinfo *si = (struct soinfo*)p;
		//printf("libname: %s\n", si->name);
		unsigned int *f = (unsigned int*)&si->symtab;
		//printf("symtab addr in mem: %x\n", f);
		//printf("symtab addr in file: %x\n", (unsigned char*)f - lnk);
		unsigned int *st = (unsigned int*) *f;
		//printf("symtab at abs-mem: %x\n", st);
		st = (unsigned int*)((unsigned int)st - 0xb0000000);
		//printf("%x\n", st);
		// 20 is the first elemet of the symbol table, it is empty therefore
		// we skip it
		st = (unsigned int*)((unsigned int)st + (unsigned int)lnk + 20);
		//printf("%x\n", *st);

		addr = *st;
		//printf("%x\n", addr);
	}
	
	munmap(lnk, buf.st_size);
	close(fp);

	return addr;
}

#ifdef DEBUG
int main(int argc, char **argv)
{
	printf("dlopen addr: %x\n", find_dlopen(argv[1]));
}
#endif
